# coding: utf-8
# Copyright (c) 2015 Fabian Barkhau <fabian.barkhau@gmail.com>
# License: MIT (see LICENSE file)


class BlockchainService(object):

    def __init__(self, testnet=False, dryrun=False):
        """TODO doc string"""
        self.testnet = testnet
        self.dryrun = dryrun

    def get_tx(self, txid):
        """TODO doc string"""
        raise NotImplementedError()

    def send_tx(self, tx):
        """TODO doc string"""
        raise NotImplementedError()

    def spendables_for_address(self, bitcoin_address):
        """TODO doc string"""
        raise NotImplementedError()

    def spendables_for_addresses(self, bitcoin_addresses):
        spendables = []
        for addr in bitcoin_addresses:
            spendables.extend(self.spendables_for_address(addr))
        return spendables





# coding: utf-8
# Copyright (c) 2015 Fabian Barkhau <fabian.barkhau@gmail.com>
# License: MIT (see LICENSE file)


from btctxstore.services.insight import Insight
from btctxstore.services.blockexplorer import BlockExplorer
from btctxstore.services.automatic import Automatic

_all = {
    "insight": Insight,
    "blockexplorer": BlockExplorer,
}

def select(name, testnet=False, dryrun=False):
    service_class = _all.get(name)
    if service_class is not None:
        return service_class(testnet=testnet, dryrun=dryrun)
    if name != "automatic":
        raise Exception("Service {0} not found!".format(name))
    return Automatic(testnet=testnet, dryrun=dryrun,
                     service_classes=_all.values())
    



# coding: utf-8
# Copyright (c) 2015 Fabian Barkhau <fabian.barkhau@gmail.com>
# License: MIT (see LICENSE file)


import random
import logging
from btctxstore.services.interface import BlockchainService


_log = logging.getLogger(__name__)


class Automatic(BlockchainService):

    def __init__(self, testnet=False, dryrun=False, service_classes=[]):
        super(Automatic, self).__init__(testnet=testnet, dryrun=dryrun)
        if not service_classes:
            raise ValueError()
        self.services = []
        for service_class in service_classes:
            self.services.append(service_class(testnet=testnet, dryrun=dryrun))

    def _select_service(self):
        return random.choice(self.services)

    def _select_other_service(self, service):
        for other_service in self.services:
            if other_service is not service:
                return other_service
        raise Exception("Not enough services!")

    def get_tx(self, txid):
        service = self._select_service()
        try:
            return service.get_tx(txid)
        except Exception as e:
            # try only once with another service
            # if two independant services fail something is wrong
            # there are also only two working services right now ...
            name = service.__class__.__name__
            msg = "Service call to {0} failed: {1}"
            _log.error(msg.format(name, repr(e)))
            other_service = self._select_other_service(service)
            return other_service.get_tx(txid)

    def send_tx(self, tx):
        service = self._select_service()
        try:
            return service.send_tx(tx)
        except Exception as e:
            # try only once with another service
            # if two independant services fail something is wrong
            # there are also only two working services right now ...
            name = service.__class__.__name__
            msg = "Service call to {0} failed: {1}"
            _log.error(msg.format(name, repr(e)))
            other_service = self._select_other_service(service)
            return other_service.send_tx(tx)

    # TODO override spendables_for_addresses in case implementation optimizes

    def spendables_for_address(self, bitcoin_address):
        service = self._select_service()
        try:
            return service.spendables_for_address(bitcoin_address)
        except Exception as e:
            # try only once with another service
            # if two independant services fail something is wrong
            # there are also only two working services right now ...
            name = service.__class__.__name__
            msg = "Service call to {0} failed: {1}"
            _log.error(msg.format(name, repr(e)))
            other_service = self._select_other_service(service)
            return other_service.spendables_for_address(bitcoin_address)



import io
import json
import logging
try:
    from urllib2 import HTTPError, urlopen
    from urllib import urlencode
except ImportError:
    from urllib.request import urlopen
    from urllib.error import HTTPError
    from urllib.parse import urlencode
from pycoin.serialize import b2h, b2h_rev, h2b, h2b_rev
from pycoin.tx import Spendable, Tx
from pycoin.convention import btc_to_satoshi
from btctxstore.services.interface import BlockchainService


_log = logging.getLogger(__name__)


class BlockExplorer(BlockchainService):

    def __init__(self, testnet=False, dryrun=False):
        super(BlockExplorer, self).__init__(testnet=testnet, dryrun=dryrun)
        if testnet:
            base_url = "https://testnet.blockexplorer.com/api"
        else:
            base_url = "https://blockexplorer.com/api"
        self.base_url = base_url

    def get_tx(self, tx_hash):
        url = "%s/rawtx/%s" % (self.base_url, b2h_rev(tx_hash))
        result = json.loads(urlopen(url).read().decode("utf8"))
        tx = Tx.tx_from_hex(result["rawtx"])
        if tx.hash() == tx_hash:
            return tx
        return None

    def send_tx(self, tx):
        if self.dryrun:
            return
        s = io.BytesIO()
        tx.stream(s)
        tx_as_hex = b2h(s.getvalue())
        data = urlencode(dict(rawtx=tx_as_hex)).encode("utf8")
        url = "%s/tx/send" % self.base_url
        try:
            d = urlopen(url, data=data).read()
            return d
        except HTTPError as ex:
            d = ex.read()
            _log.error(repr(ex))

    # TODO override spendables_for_addresses
    # api support multiple addresses per query

    def spendables_for_address(self, bitcoin_address):
        url = "{0}/addr/{1}/utxo".format(self.base_url, bitcoin_address)
        result = json.loads(urlopen(url).read().decode("utf8"))
        spendables = []
        for utxo in result:
            value = btc_to_satoshi(str(utxo["amount"]))
            prev_index = utxo["vout"]
            prev_hash = h2b_rev(utxo["txid"])
            script = h2b(utxo["scriptPubKey"])
            spendable = Spendable(value, script, prev_hash, prev_index)
            spendables.append(spendable)
        return spendables



#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2015 Fabian Barkhau <fabian.barkhau@gmail.com>
# License: MIT (see LICENSE file)


from __future__ import print_function
from __future__ import unicode_literals
import unittest
from btctxstore import exceptions
from btctxstore import BtcTxStore


S_HWIF = "tprv8ZgxMBicQKsPeFZuyoMy197pX8iJeTasLVYEeb68cj489fmCz7kxP8vAfJp3rWebDT3nLvDsBaz3CqnUfdhr1wbjmhUNwK5Hf7dN2Er9btD"
B_HWIF = b"tprv8ZgxMBicQKsPeFZuyoMy197pX8iJeTasLVYEeb68cj489fmCz7kxP8vAfJp3rWebDT3nLvDsBaz3CqnUfdhr1wbjmhUNwK5Hf7dN2Er9btD"
U_HWIF = u"tprv8ZgxMBicQKsPeFZuyoMy197pX8iJeTasLVYEeb68cj489fmCz7kxP8vAfJp3rWebDT3nLvDsBaz3CqnUfdhr1wbjmhUNwK5Hf7dN2Er9btD"


class TestValidateWallet(unittest.TestCase):

    def setUp(self):
        self.testnet_api = BtcTxStore(dryrun=True, testnet=True)
        self.mainnet_api = BtcTxStore(dryrun=True, testnet=False)
        self.testnet_wallet = self.testnet_api.create_wallet()
        self.mainnet_wallet = self.mainnet_api.create_wallet()
        self.testnet_key = self.testnet_api.get_key(self.testnet_wallet)
        self.mainnet_key = self.mainnet_api.get_key(self.mainnet_wallet)

    def test_checks_networks(self):
        self.assertTrue(self.testnet_api.validate_wallet(self.testnet_wallet))
        self.assertTrue(self.mainnet_api.validate_wallet(self.mainnet_wallet))
        self.assertFalse(self.testnet_api.validate_wallet(self.mainnet_wallet))
        self.assertFalse(self.mainnet_api.validate_wallet(self.testnet_wallet))

    def test_doesnt_validate_keys(self):
        self.assertFalse(self.testnet_api.validate_wallet(self.testnet_key))
        self.assertFalse(self.mainnet_api.validate_wallet(self.testnet_key))
        self.assertFalse(self.testnet_api.validate_wallet(self.mainnet_key))
        self.assertFalse(self.mainnet_api.validate_wallet(self.mainnet_key))

    def test_correct_types(self):
        self.assertTrue(self.testnet_api.validate_wallet(S_HWIF))
        self.assertTrue(self.testnet_api.validate_wallet(B_HWIF))
        self.assertTrue(self.testnet_api.validate_wallet(U_HWIF))


if __name__ == '__main__':
    unittest.main()
